/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * This program is written in GAP-L as developed by Georg Sauthoff at Bielefeld
 * University. GAP-L, the Bellman's Gap Compiler and other language features are
 * used in accordance to the GPL v3+.
 *
 * Bellman's Gap can be found at http://www.gapc.eu
 */

/**
 * This implements the global sequence alignment on simple character sequences.
 * Every character is interpreted as a discrete symbol.
 *
 * @author =  Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
//just some preamble definitions

import align_cost
import twoline_Rope

input < raw, raw >


type Rope = extern
type twoline_Rope = extern

/*
 * Signature that defines the operations allowed.
 */
signature sig_alignment(alphabet, answer) {
  answer del(<alphabet,void>,answer);
  answer ins(<void,alphabet>,answer);
  answer rep(<alphabet,alphabet>,answer);
  answer nil(<void,void>);
  choice [answer] h([answer]);
}

/*
 * Algebra section.
 */
algebra enum auto enum ;
algebra count auto count;

algebra alg_strict implements sig_alignment(alphabet = char, answer = double){

	double del(<char left, void>, double in){
		return in + del_cost(left);
	}
	
	double ins(<void, char right>, double in){
		return in + ins_cost(right);
	}
	
	double rep(<char left, char right>, double in){
		return in + rep_cost(left, right);
	}
	
	double nil(<void, void>){
		return 0;
	}
	
	choice [double] h([double] l){
		return list(minimum(l));
	}
}

algebra alg_print implements sig_alignment(alphabet = char, answer = twoline_Rope){

	twoline_Rope del(<char left,void>, twoline_Rope in){
		twoline_Rope out;
		append(out.top, left);
		append(out.top, in.top);
		append(out.bottom, '-');
		append(out.bottom, in.bottom);
		return out;
	}
	
	twoline_Rope ins(<void,char right>, twoline_Rope in){
		twoline_Rope out;
		append(out.top, '-');
		append(out.top, in.top);
		append(out.bottom, right);
		append(out.bottom, in.bottom);
		return out;
	}
	
	twoline_Rope rep(<char left, char right>, twoline_Rope in){
		twoline_Rope out;
		append(out.top, left);
		append(out.top, in.top);
		append(out.bottom, right);
		append(out.bottom, in.bottom);
		return out;
	}
	
	twoline_Rope nil(<void,void>){
		twoline_Rope out;
		return out;
	}
	
	choice [twoline_Rope] h([twoline_Rope] l){
		return l;
	}
}

/*
 * Note that this grammar is _not_ equivalent to our Java implementation of
 * global alignment. Here we prevent semantic ambiguity by enforcing that
 * deletions come before insertions. Thus we prevent situations where the input
 * "a" and "a"
 * can be aligned as ins(del('a',nil),'a') as well as del('a',ins(nil,'a'))
 */
grammar gra_alignment uses sig_alignment(axiom = ALI)
{

	ALI = del(<CHAR,EMPTY>, ALI) | ins(<EMPTY,CHAR>, INS) | rep(<CHAR,CHAR>, ALI) | nil(<EMPTY,EMPTY>) #h;

	INS = ins(<EMPTY,CHAR>, INS) | rep(<CHAR,CHAR>, ALI) | nil(<EMPTY,EMPTY>) #h;
}


/*
 * This defines compilable instances of this program.
 * The command line usage (if gapc is installed):
 *
 * gapc -i <instance name> -o global_alignment.cc global_alignment.gap
 *
 */
instance print	= gra_alignment(alg_print);
instance score	= gra_alignment(alg_strict);
instance path	= gra_alignment(alg_strict * alg_print);
instance debug	= gra_alignment(alg_print * alg_strict);
