/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment.comparators;

import de.unibi.citec.tcs.alignment.sequence.Alphabet;
import java.util.Arrays;

/**
 * This is a matrix of parameters specifying costs to replace one symbol of an
 * alphabet with another one. If you want to use this for alignment, you have to
 * specify deletion and insertion costs, too. If you furthermore want to use
 * this for local alignment you have to specify skip costs as well.
 *
 * Please note that this matrix supports an individual scoring scheme for the
 * costs of each combination of symbols in the given alphabet.
 *
 * More formally: For each combination of symbols a,b in Sigma u {_,skip} you
 * have to specify
 *
 * lambda(a,b), where lambda(a,b) is stored at point [i][j] in the matrix, where
 * i is the index of a in the alphabet and j is the index of b in the alphabet
 * (see also the Alphabet documentation).
 *
 * deletion costs
 *
 * lambda(a,_)
 *
 * are stored at point [i][|Sigma|] and insertion costs
 *
 * lambda(_,b)
 *
 * are stored at point [|Sigma|][j]. Accordingly skip costs
 *
 * lambda(a,skip)
 *
 * are stored at point [i][|Sigma|+1] and [|Sigma|+1][j] respectively.
 *
 *
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class ReplacementCosts {

	private final Alphabet alphabet;
	private final double[][] costMatrix;
	public static final double MISMATCHDEFAULT = 1;
	public static final double GAPDEFAULT = 1;
	public static final double SKIPDEFAULT = 0.5;

	/**
	 * Initializes an empty replacement score matrix with sensible defaults: The
	 * diagonal is set to 0 (it costs nothing to replace a symbol by itself),
	 * every other replacement is set to MISMATCHDEFAULT, gap costs are set to
	 * GAPDEFAULT and skip costs to SKIPDEFAULT.
	 *
	 * @param alphabet the alphabet that defines the indexing for this
	 * replacement cost matrix.
	 * @param supportsGaps set to true if you also want to supply gap costs to
	 * make this cost matrix usable for alignments.
	 * @param supportsSkips set to true if you also want to supply skip costs to
	 * make this cost matrix usable for local alignments. Please note that
	 * supportGaps has to be true in this case, too.
	 */
	public ReplacementCosts(final Alphabet alphabet, final boolean supportsGaps,
			final boolean supportsSkips) {
		this.alphabet = alphabet;
		int additionalLines = 0;
		if (supportsGaps) {
			additionalLines++;
			if (supportsSkips) {
				additionalLines++;
			}
		}
		this.costMatrix = new double[alphabet.size() + additionalLines][alphabet.size()
				+ additionalLines];
		//the diagonal ist initialized instantly.
		//the other replacement costs are set here.
		for (int i = 0; i < alphabet.size(); i++) {
			for (int j = 0; j < alphabet.size(); j++) {
				if (i != j) {
					costMatrix[i][j] = MISMATCHDEFAULT;
				}
			}
		}
		//then we set gap and skip costs.
		for (int i = 0; i < alphabet.size(); i++) {
			costMatrix[i][alphabet.size()] = GAPDEFAULT;
			costMatrix[i][alphabet.size() + 1] = SKIPDEFAULT;
			costMatrix[alphabet.size()][i] = GAPDEFAULT;
			costMatrix[alphabet.size() + 1][i] = SKIPDEFAULT;
		}
	}

	public ReplacementCosts(Alphabet alphabet, double[][] costMatrix) {
		this.alphabet = alphabet;
		if (costMatrix.length < alphabet.size()) {
			throw new UnsupportedOperationException(
					"The given matrix does not specify all necessary replacement costs!");
		}
		if (costMatrix.length > alphabet.size() + 2) {
			throw new UnsupportedOperationException(
					"The given matrix is too large! Only costs for replacements, deletions, insertions and skips may be specified!");
		}
		for (int i = 0; i < costMatrix.length; i++) {
			if (costMatrix[i].length != costMatrix.length) {
				throw new UnsupportedOperationException("The given matrix is not quadratic!");
			}
		}
		this.costMatrix = costMatrix;
	}

	/**
	 *
	 * @return true if this matrix specifies gap costs.
	 */
	public boolean isGapAlignmentMatrix() {
		return this.costMatrix.length > alphabet.size();
	}

	/**
	 *
	 * @return true if this matrix specifies skip costs.
	 */
	public boolean isLocalAlignmentMatrix() {
		return this.costMatrix.length > alphabet.size() + 1;
	}

	/**
	 * Returns the cost matrix itself. See the class documentation for a more
	 * detailed description.
	 *
	 * @return the cost matrix itself.
	 */
	public double[][] getCostMatrix() {
		return costMatrix;
	}

	/**
	 * Sets the cost matrix itself. See the class documentation for a more
	 * detailed description.
	 *
	 * Please note that this matrix is copied into the initialized matrix. For
	 * big matrices you should consider getting the current matrix and setting
	 * single values there.
	 *
	 * @param costMatrix the cost matrix itself.
	 */
	public void setCostMatrix(double[][] costMatrix) {
		//check the dimensions.
		if (costMatrix.length != this.costMatrix.length) {
			throw new RuntimeException("Expected length " + this.costMatrix.length
					+ " but was " + costMatrix.length);
		}
		for (int i = 0; i < costMatrix.length; i++) {
			if (costMatrix[i].length != this.costMatrix[i].length) {
				throw new RuntimeException("Expected length " + this.costMatrix[i].length
						+ " but was " + costMatrix[i].length);
			}
			System.arraycopy(costMatrix[i], 0, this.costMatrix[i], 0, costMatrix[i].length);
		}
	}

	/**
	 * Returns the whole matrix of replacement costs without gap and skip costs.
	 * Please note that in contrast to "getCostMatrix" this is a copy of the
	 * array.
	 *
	 * @return the whole matrix of replacement costs without gap and skip costs.
	 */
	public double[][] getReplacementCosts() {
		final double[][] replacementMatrix = new double[alphabet.size()][alphabet.size()];
		for (int i = 0; i < replacementMatrix.length; i++) {
			System.arraycopy(costMatrix[i], 0, replacementMatrix[i], 0, replacementMatrix.length);
		}
		return replacementMatrix;

	}

	/**
	 * Sets the whole matrix of replacement costs without gap and skip costs.
	 *
	 * @param replacementCosts the whole matrix of replacement costs without gap
	 * and skip costs.
	 */
	public void setReplacementCosts(double[][] replacementCosts) {
		if (replacementCosts.length != alphabet.size()) {
			throw new RuntimeException("Expected length " + alphabet.size()
					+ " but was " + replacementCosts.length);
		}

		for (int i = 0; i < alphabet.size(); i++) {
			if (replacementCosts[i].length != alphabet.size()) {
				throw new RuntimeException("Expected length " + alphabet.size()
						+ " but was " + replacementCosts[i].length);
			}
			System.arraycopy(replacementCosts[i], 0, costMatrix[i], 0, alphabet.size());
		}
	}

	public double getReplacementCost(String a, String b) {
		final int aIdx = alphabet.getSymbolIndex(a);
		if (aIdx == -1) {
			throw new RuntimeException("Symbol " + a + " is not specified in the alphabet!");
		}
		final int bIdx = alphabet.getSymbolIndex(b);
		if (bIdx == -1) {
			throw new RuntimeException("Symbol " + b + " is not specified in the alphabet!");
		}
		return costMatrix[aIdx][bIdx];
	}

	public void setReplacementCost(String a, String b, double cost) {
		final int aIdx = alphabet.getSymbolIndex(a);
		if (aIdx == -1) {
			throw new RuntimeException("Symbol " + a + " is not specified in the alphabet!");
		}
		final int bIdx = alphabet.getSymbolIndex(b);
		if (bIdx == -1) {
			throw new RuntimeException("Symbol " + b + " is not specified in the alphabet!");
		}
		costMatrix[aIdx][bIdx] = cost;
	}

	/**
	 * Returns all deletion costs without any other costs. Please note that in
	 * contrast to "getCostMatrix" this is a copy of the array.
	 *
	 * @return all deletion costs without any other costs.
	 */
	public double[] getDeletionCosts() {
		if (!isGapAlignmentMatrix()) {
			throw new RuntimeException("This matrix does not support gap costs!");
		}
		final double[] delCosts = new double[alphabet.size()];
		for (int i = 0; i < alphabet.size(); i++) {
			delCosts[i] = costMatrix[i][alphabet.size()];
		}
		return delCosts;
	}

	/**
	 * Sets all deletion costs without any other costs.
	 *
	 * @param delCosts all deletion costs without any other costs.
	 */
	public void setDeletionCosts(double[] delCosts) {
		if (!isGapAlignmentMatrix()) {
			throw new RuntimeException("This matrix does not support gap costs!");
		}
		if (delCosts.length != alphabet.size()) {
			throw new RuntimeException("Expected length " + alphabet.size()
					+ " but was " + delCosts.length);
		}

		for (int i = 0; i < alphabet.size(); i++) {
			costMatrix[i][alphabet.size()] = delCosts[i];
		}
	}

	public double getDeletionCost(String a) {
		if (isGapAlignmentMatrix()) {
			final int aIdx = alphabet.getSymbolIndex(a);
			if (aIdx == -1) {
				throw new RuntimeException("Symbol " + a + " is not specified in the alphabet!");
			}
			return costMatrix[aIdx][alphabet.size()];
		} else {
			throw new RuntimeException("This matrix does not support gap costs!");
		}
	}

	public void setDeletionCost(String a, double cost) {
		if (isGapAlignmentMatrix()) {
			final int aIdx = alphabet.getSymbolIndex(a);
			if (aIdx == -1) {
				throw new RuntimeException("Symbol " + a + " is not specified in the alphabet!");
			}
			costMatrix[aIdx][alphabet.size()] = cost;
		} else {
			throw new RuntimeException("This matrix does not support gap costs!");
		}
	}

	/**
	 * Returns all insertion costs without any other costs. Please note that in
	 * contrast to "getCostMatrix" this is a copy of the array.
	 *
	 * @return all insertion costs without any other costs.
	 */
	public double[] getInsertionCosts() {
		if (!isGapAlignmentMatrix()) {
			throw new RuntimeException("This matrix does not support gap costs!");
		}
		final double[] insCosts = new double[alphabet.size()];
		System.arraycopy(costMatrix[alphabet.size()], 0, insCosts, 0, alphabet.size());
		return insCosts;
	}

	/**
	 * Sets all insertion costs without any other costs.
	 *
	 * @param insCosts all insertion costs without any other costs.
	 */
	public void setInsertionCosts(double[] insCosts) {
		if (!isGapAlignmentMatrix()) {
			throw new RuntimeException("This matrix does not support gap costs!");
		}
		if (insCosts.length != alphabet.size()) {
			throw new RuntimeException("Expected length " + alphabet.size()
					+ " but was " + insCosts.length);
		}
		System.arraycopy(insCosts, 0, costMatrix[alphabet.size()], 0, alphabet.size());
	}

	public double getInsertionCost(String b) {
		if (isGapAlignmentMatrix()) {
			final int bIdx = alphabet.getSymbolIndex(b);
			if (bIdx == -1) {
				throw new RuntimeException("Symbol " + b + " is not specified in the alphabet!");
			}
			return costMatrix[alphabet.size()][bIdx];
		} else {
			throw new RuntimeException("This matrix does not support gap costs!");
		}
	}

	public void setInsertionCost(String b, double cost) {
		if (isGapAlignmentMatrix()) {
			final int bIdx = alphabet.getSymbolIndex(b);
			if (bIdx == -1) {
				throw new RuntimeException("Symbol " + b + " is not specified in the alphabet!");
			}
			costMatrix[alphabet.size()][bIdx] = cost;
		} else {
			throw new RuntimeException("This matrix does not support gap costs!");
		}
	}

	/**
	 * Returns all skip deletion costs without any other costs. Please note that
	 * in contrast to "getCostMatrix" this is a copy of the array.
	 *
	 * @return all skip deletion costs without any other costs.
	 */
	public double[] getSkipDeletionCosts() {
		if (!isGapAlignmentMatrix()) {
			throw new RuntimeException("This matrix does not support skip costs!");
		}
		final double[] delCosts = new double[alphabet.size()];
		for (int i = 0; i < alphabet.size(); i++) {
			delCosts[i] = costMatrix[i][alphabet.size() + 1];
		}
		return delCosts;
	}

	/**
	 * Sets all skip deletion costs without any other costs.
	 *
	 * @param delCosts all skip deletion costs without any other costs.
	 */
	public void setSkipDeletionCosts(double[] delCosts) {
		if (!isLocalAlignmentMatrix()) {
			throw new RuntimeException("This matrix does not support skip costs!");
		}
		if (delCosts.length != alphabet.size()) {
			throw new RuntimeException("Expected length " + alphabet.size()
					+ " but was " + delCosts.length);
		}

		for (int i = 0; i < alphabet.size(); i++) {
			costMatrix[i][alphabet.size() + 1] = delCosts[i];
		}
	}

	public double getSkipDeletionCost(String a) {
		if (isLocalAlignmentMatrix()) {
			final int aIdx = alphabet.getSymbolIndex(a);
			if (aIdx == -1) {
				throw new RuntimeException("Symbol " + a + " is not specified in the alphabet!");
			}
			return costMatrix[aIdx][alphabet.size() + 1];
		} else {
			throw new RuntimeException("This matrix does not support skip costs!");
		}
	}

	public void setSkipDeletionCost(String a, double cost) {
		if (isLocalAlignmentMatrix()) {
			final int aIdx = alphabet.getSymbolIndex(a);
			if (aIdx == -1) {
				throw new RuntimeException("Symbol " + a + " is not specified in the alphabet!");
			}
			costMatrix[aIdx][alphabet.size() + 1] = cost;
		} else {
			throw new RuntimeException("This matrix does not support skip costs!");
		}
	}

	/**
	 * Returns all skip insertion costs without any other costs. Please note
	 * that in contrast to "getCostMatrix" this is a copy of the array.
	 *
	 * @return all skip insertion costs without any other costs.
	 */
	public double[] getSkipInsertionCosts() {
		if (!isGapAlignmentMatrix()) {
			throw new RuntimeException("This matrix does not support skip costs!");
		}
		final double[] insCosts = new double[alphabet.size()];
		System.arraycopy(costMatrix[alphabet.size() + 1], 0, insCosts, 0, alphabet.size());
		return insCosts;
	}

	/**
	 * Sets all skip insertion costs without any other costs.
	 *
	 * @param insCosts all skip insertion costs without any other costs.
	 */
	public void setSkipInsertionCosts(double[] insCosts) {
		if (!isGapAlignmentMatrix()) {
			throw new RuntimeException("This matrix does not support skip costs!");
		}
		if (insCosts.length != alphabet.size()) {
			throw new RuntimeException("Expected length " + alphabet.size()
					+ " but was " + insCosts.length);
		}
		System.arraycopy(insCosts, 0, costMatrix[alphabet.size() + 1], 0, alphabet.size());
	}

	public double getSkipInsertionCost(String b) {
		if (isLocalAlignmentMatrix()) {
			final int bIdx = alphabet.getSymbolIndex(b);
			if (bIdx == -1) {
				throw new RuntimeException("Symbol " + b + " is not specified in the alphabet!");
			}
			return costMatrix[alphabet.size() + 1][bIdx];
		} else {
			throw new RuntimeException("This matrix does not support skip costs!");
		}
	}

	public void setSkipInsertionCost(String b, double cost) {
		if (isLocalAlignmentMatrix()) {
			final int bIdx = alphabet.getSymbolIndex(b);
			if (bIdx == -1) {
				throw new RuntimeException("Symbol " + b + " is not specified in the alphabet!");
			}
			costMatrix[alphabet.size() + 1][bIdx] = cost;
		} else {
			throw new RuntimeException("This matrix does not support skip costs!");
		}
	}

	public Alphabet getAlphabet() {
		return alphabet;
	}

	@Override
	public int hashCode() {
		int hash = 7;
		hash = 17 * hash + Arrays.deepHashCode(this.costMatrix);
		return hash;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final ReplacementCosts other = (ReplacementCosts) obj;
		if (!Arrays.deepEquals(this.costMatrix, other.costMatrix)) {
			return false;
		}
		return true;
	}
}
