/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.comparators.ReplacementComparator;
import de.unibi.citec.tcs.alignment.comparators.TrivialEditComparator;
import de.unibi.citec.tcs.alignment.sequence.Alphabet;
import de.unibi.citec.tcs.alignment.sequence.KeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.Node;
import de.unibi.citec.tcs.alignment.sequence.NodeSpecification;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.sequence.SymbolicKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.SymbolicValue;
import de.unibi.citec.tcs.alignment.sequence.VectorialKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.VectorialValue;
import de.unibi.citec.tcs.alignment.wrappers.StringEditDistance;
import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.util.Random;
import org.junit.After;
import org.junit.AfterClass;
import static org.junit.Assert.*;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

/**
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class ParallelDerivativeEngineTest {

	private static AlignmentPath[] paths;

	public ParallelDerivativeEngineTest() {
	}

	@BeforeClass
	public static void setUpClass() {
		/*
		 * we create two sequences with two keywords and one node each. One
		 * keyword has no differences at all, while the other one is
		 * discriminative.
		 */
		final KeywordSpecification[] keywordSpecs = new KeywordSpecification[2];
		final Alphabet alpha = new Alphabet("a|b");
		keywordSpecs[0] = new SymbolicKeywordSpecification(alpha, "discr");
		keywordSpecs[1] = new VectorialKeywordSpecification(1, "nondiscr");

		final NodeSpecification nodeSpec = new NodeSpecification(keywordSpecs);

		final Sequence[] seqs = new Sequence[2];
		seqs[0] = new Sequence(nodeSpec);
		Node n = new Node(seqs[0]);
		n.setValue("discr", new SymbolicValue(alpha, "a"));
		n.setValue("nondiscr", new VectorialValue(8));
		seqs[0].getNodes().add(n);

		seqs[1] = new Sequence(nodeSpec);
		n = new Node(seqs[1]);
		n.setValue("discr", new SymbolicValue(alpha, "b"));
		n.setValue("nondiscr", new VectorialValue(8));
		seqs[1].getNodes().add(n);

		final AlignmentSpecification alignSpec = new AlignmentSpecification(nodeSpec);
		final StrictAlignmentFullAlgorithm algo = new StrictAlignmentFullAlgorithm(alignSpec);

		paths = new AlignmentPath[2];
		paths[0] = algo.calculateAlignment(seqs[0], seqs[1]);
		paths[1] = algo.calculateAlignment(seqs[1], seqs[0]);
	}

	@AfterClass
	public static void tearDownClass() {
	}

	@Before
	public void setUp() {
	}

	@After
	public void tearDown() {
	}

	/**
	 * Test of calculateWeightDerivatives method, of class
	 * ParallelDerivativeEngine.
	 */
	@Test
	public void testCalculateWeightDerivatives() {
		final ParallelDerivativeEngine instance = new ParallelDerivativeEngine(paths);
		instance.setVerbose(false);
		final double[][] weightDerivatives = instance.calculateWeightDerivatives();
		assertEquals(paths.length, weightDerivatives.length);
		for (int i = 0; i < paths.length; i++) {
			/*
			 * we expect zero derivative for the non-discriminatory keyword and
			 * a 1-derivative for the discriminatory one..
			 */
			final double[] currentDeriv = weightDerivatives[i];
			assertEquals(2, currentDeriv.length);
			assertEquals(1, currentDeriv[0], 0);
			assertEquals(0, currentDeriv[1], 0);
		}
	}

	/**
	 * Test of calculateComparatorDerivatives method, of class
	 * ParallelDerivativeEngine.
	 */
	@Test
	public void testCalculateComparatorDerivatives() {
		final ParallelDerivativeEngine instance = new ParallelDerivativeEngine(paths);
		instance.setVerbose(false);
		final double[][][] derivatives = instance.calculateComparatorDerivatives(
				(ReplacementComparator) paths[0].getSpecificaton().getComparator(0),
				"discr");
		assertEquals(paths.length, derivatives.length);
		/*
		 * For the first path we expect a 0.5-derivative (because the weight is
		 * 0.5) for the a->b replacement and otherwise zero.
		 */
		double[][] currentDeriv = derivatives[0];
		for (int i = 0; i < 2; i++) {
			for (int j = 0; j < 2; j++) {
				if (i == 0 && j == 1) {
					assertEquals(0.5, currentDeriv[i][j], 0);
				} else {
					assertEquals(0, currentDeriv[i][j], 0);
				}
			}
		}
		/*
		 * Conversely for the snd path we expect a 0.5-derivative for the b->a
		 * replacement and otherwise zero.
		 */
		currentDeriv = derivatives[1];
		for (int i = 0; i < 2; i++) {
			for (int j = 0; j < 2; j++) {
				if (i == 1 && j == 0) {
					assertEquals(0.5, currentDeriv[i][j], 0);
				} else {
					assertEquals(0, currentDeriv[i][j], 0);
				}
			}
		}
	}

	/**
	 * Test of calculate method, of class ParallelProcessingEngine.
	 */
	@Test
	public void testCalculationSpeedUp() {

		//generate longer test sequences.
		final int seqNum = 50;
		final int seqLength = 100;
		final int alphabetSize = 5;
		final Sequence[] sequences = new Sequence[seqNum];
		final Random rand = new Random();
		for (int i = 0; i < seqNum; i++) {
			final char[] seqIndices = new char[seqLength];
			for (int j = 0; j < seqLength; j++) {
				seqIndices[j] = (char) (65 + rand.nextInt(alphabetSize));
			}
			sequences[i] = new Sequence(new String(seqIndices));
		}

		//set up algorithm.
		final SoftAlignmentFullAlgorithm algo = new SoftAlignmentFullAlgorithm(
				new AlignmentSpecification(sequences[0]));

		final ReplacementComparator comp = (ReplacementComparator) algo.getSpecification().
				getComparator(0);

		//calculate some alignment paths.
		final ParallelProcessingEngine<SoftPathModel> calcEngine
				= new ParallelProcessingEngine<SoftPathModel>(algo, sequences);
		calcEngine.setVerbose(false);
		final boolean[][] calcSetup = new boolean[seqNum][seqNum];
		for (int i = 0; i < seqNum; i++) {
			calcSetup[0][i] = true;
		}
		calcEngine.setDoCalculation(calcSetup);
		calcEngine.calculate();
		final SoftPathModel[][] paths = calcEngine.getResultMatrix();

		//set up instance
		final ParallelDerivativeEngine instance = new ParallelDerivativeEngine(paths);
		instance.setVerbose(false);

		//do a sequential computation.
		final double[][][] benchmark = new double[seqNum][][];
		long start = System.nanoTime();
		for (int i = 0; i < seqNum; i++) {
			benchmark[i] = paths[0][i].calculateParameterDerivative(
					comp, Sequence.DEFAULTKEYWORD);

		}
		long end = System.nanoTime();
		long sequentialTime = end - start;
		//do a parallel computation
		start = System.nanoTime();
		final double[][][] actual = instance.calculateComparatorDerivatives(
				comp, Sequence.DEFAULTKEYWORD);
		end = System.nanoTime();
		long parallelTime = end - start;

		System.out.println("sequential:" + sequentialTime / 1000000 + " ms");
		System.out.println("parallel:" + parallelTime / 1000000 + " ms");
		//we expect at least a speedup for parallel computation.
		assertTrue(parallelTime < sequentialTime);
		//check the results.
		assertEquals(seqNum, benchmark.length);
		assertEquals(seqNum, actual.length);
		for (int i = 0; i < seqNum; i++) {
			assertEquals(alphabetSize + 2, benchmark[i].length);
			assertEquals(alphabetSize + 2, actual[i].length);
			for (int j = 0; j < alphabetSize; j++) {
				assertEquals(alphabetSize + 2, benchmark[i][j].length);
				assertEquals(alphabetSize + 2, actual[i][j].length);
				for (int k = 0; k < alphabetSize; k++) {
					assertEquals(benchmark[i][j][k], actual[i][j][k], 0);
				}
			}
		}
	}

	@Test
	public void testExceptionHandling() throws UnsupportedEncodingException {
		//create a valid alignment path.
		final Sequence[] seqs = StringEditDistance.toSequences(new String[]{"aaa", "aba"});
		final AlignmentSpecification alignSpec = StringEditDistance.setUpSpecification(seqs);
		final StrictAlignmentFullAlgorithm algo = new StrictAlignmentFullAlgorithm(alignSpec);
		final AlignmentPath path = algo.calculateAlignment(seqs[0], seqs[1]);
		//now we try to calculate the parameter derivative for an entirely unrelated comparator.
		final TrivialEditComparator comp = new TrivialEditComparator();
		//... and we expect an exception.
		boolean exception = false;
		try {
			path.calculateParameterDerivative(comp, "content");
		} catch (UnsupportedOperationException ex) {
			assertEquals("The given comparator was not used for the given keyword!", ex.getMessage());
			exception = true;
		}
		assertTrue("We expected an exception!", exception);
		//the same exception should be thrown if we try that within a
		//ParallelDerivativeEngine
		final ParallelDerivativeEngine instance = new ParallelDerivativeEngine(
				new AlignmentDerivativeAlgorithm[]{path});
		exception = false;
		try {
			instance.calculateComparatorDerivatives(comp, "content");
		} catch (UnsupportedOperationException ex) {
			assertEquals("The given comparator was not used for the given keyword!", ex.getMessage());
			exception = true;
		}
		assertTrue("We expected an exception!", exception);
	}

}
