/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.comparators.ReplacementComparator;
import de.unibi.citec.tcs.alignment.sequence.Alphabet;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import junit.framework.TestCase;

/**
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class SoftAlignmentFullAlgorithmTest extends TestCase {

	public SoftAlignmentFullAlgorithmTest(String testName) {
		super(testName);
	}

	@Override
	protected void setUp() throws Exception {
		super.setUp();
	}

	@Override
	protected void tearDown() throws Exception {
		super.tearDown();
	}

	/**
	 * Test of calculateSoftAlignment method, of class SoftAlignmentAlgorithm.
	 */
	public void testPerformance() {

		//align a...a and a...b
		final int length = 1000;

		final StringBuilder aBuilder = new StringBuilder();
		final StringBuilder bBuilder = new StringBuilder();

		aBuilder.append('b');
		bBuilder.append('b');
		for (int i = 1; i < length - 1; i++) {
			aBuilder.append('a');
			bBuilder.append('a');
		}
		aBuilder.append('a');
		bBuilder.append('b');

		final Sequence a = new Sequence(aBuilder.toString());
		final Sequence b = new Sequence(bBuilder.toString());
		//initialize actual instance.
		final SoftAlignmentFullAlgorithm instance
				= new SoftAlignmentFullAlgorithm(new AlignmentSpecification(a));
		instance.setBeta(10);

		final double result = instance.calculateAlignment(a, b).getScore();
		//the result should be very close to zero because we normalize using the length.
		assertEquals(0, result, 0.01);
	}

	/**
	 * Test of calculateSoftAlignment method, of class SoftAlignmentAlgorithm.
	 */
	public void testCalculateSoftAlignment() {
		final Sequence a = new Sequence("ab");
		final Sequence b = new Sequence("aab");
		final SoftAlignmentScoreAlgorithm baseline
				= new SoftAlignmentScoreAlgorithm(new AlignmentSpecification(a));
		final SoftAlignmentFullAlgorithm instance
				= new SoftAlignmentFullAlgorithm(baseline.getSpecification());
		double optimum = ((ReplacementComparator) baseline.getSpecification().getComparator(0)).getCostMatrix().getInsertionCost("b") / 5;

		//check if the result is bigger than the optimum but converges against
		//the optimum for higher beta.
		SoftPathModel lastResult = null;
		for (double beta = 0; beta < 20; beta++) {
			baseline.setBeta(beta);
			instance.setBeta(beta);
			final double expected = baseline.calculateAlignment(a, b);
			final SoftPathModel actual = instance.calculateAlignment(a, b);
			assertEquals(expected, actual.getScore(), 1E-6);
			lastResult = actual;
		}
		assertEquals(optimum, lastResult.getScore(), 1E-4);
		//check matrix size.
		assertEquals(a.getNodes().size() + 1, lastResult.getPathMatrix().length);
		for (int i = 0; i <= a.getNodes().size(); i++) {
			assertEquals(b.getNodes().size() + 1, lastResult.getPathMatrix()[i].length);
		}

		//use Strict alignment for this high beta value as comparison.
		final StrictAlignmentFullAlgorithm strict = new StrictAlignmentFullAlgorithm(baseline.getSpecification());
		final AlignmentPath alignment = strict.calculateAlignment(a, b);
		System.out.println(alignment.toString());

		/*
		 * check probabilities.
		 *
		 * We have two different possible paths: first insertion, then two
		 * replacements or replacement, insertion, replacement. So we check both
		 * paths.
		 */
		//first path first step
		assertEquals(1, lastResult.getPathMatrix()[1][1].getRepProb(), 1E-4);
		//second path first step
		assertEquals(1, lastResult.getPathMatrix()[0][1].getInsProb(), 1E-4);
		//this is where the paths split.
		assertEquals(0.5, lastResult.getPathMatrix()[1][2].getInsProb(), 1E-4);
		assertEquals(0.5, lastResult.getPathMatrix()[1][2].getRepProb(), 1E-4);
		//and in the last step both paths are equal.
		assertEquals(1, lastResult.getPathMatrix()[2][3].getRepProb(), 1E-4);
		//and check the score again.
		assertEquals(lastResult.getScore(), lastResult.getPathMatrix()[2][3].getSoftmin() / 5, 1E-4);
	}

}
