/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.comparators.ReplacementComparator;
import de.unibi.citec.tcs.alignment.sequence.Alphabet;
import de.unibi.citec.tcs.alignment.sequence.KeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.Node;
import de.unibi.citec.tcs.alignment.sequence.NodeSpecification;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.sequence.StringKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.StringValue;
import de.unibi.citec.tcs.alignment.sequence.SymbolicKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.SymbolicValue;
import de.unibi.citec.tcs.alignment.sequence.VectorialKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.VectorialValue;
import java.util.Arrays;
import static junit.framework.Assert.assertEquals;
import junit.framework.TestCase;

/**
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class StrictAlignmentFullAlgorithmTest extends TestCase {

	public StrictAlignmentFullAlgorithmTest(String testName) {
		super(testName);
	}

	@Override
	protected void setUp() throws Exception {
		super.setUp();
	}

	@Override
	protected void tearDown() throws Exception {
		super.tearDown();
	}

	/**
	 * Test of calculateAlignment method, of class StrictAlignmentAlgorithm.
	 */
	public void testAlignmentPerformance() {

		//align a...a and a...b
		final int length = 1000;

		final StringBuilder aBuilder = new StringBuilder();
		final StringBuilder bBuilder = new StringBuilder();

		aBuilder.append('b');
		bBuilder.append('b');
		for (int i = 1; i < length - 1; i++) {
			aBuilder.append('a');
			bBuilder.append('a');
		}
		aBuilder.append('a');
		bBuilder.append('b');

		final Sequence a = new Sequence(aBuilder.toString());
		final Sequence b = new Sequence(bBuilder.toString());

		//initialize actual instance.
		final StrictAlignmentFullAlgorithm instance
				= new StrictAlignmentFullAlgorithm(new AlignmentSpecification(a));

		double expResult = 5E-4;
		final long start = System.nanoTime();
		final AlignmentPath path = instance.calculateAlignment(a, b);
		final double time = (double) (System.nanoTime() - start) / 1E9;
		assertEquals(expResult, path.getScore(), 0);
	}

	public void testCalculateAlignment() {
		/*
		 * This is a test for an alignment of full blown sequences. Our test
		 * sequences here represent very simple java programs.
		 */
		//specify the keywords.
		final KeywordSpecification[] keywordSpecs = new KeywordSpecification[3];
		//first the node type.
		final Alphabet nodeTypes = new Alphabet(new String[]{"class", "method"});
		keywordSpecs[0] = new SymbolicKeywordSpecification(nodeTypes, "type");
		//then the name
		keywordSpecs[1] = new StringKeywordSpecification("name");
		//then the code position (only start line and end line)
		keywordSpecs[2] = new VectorialKeywordSpecification(2, "codePosition");
		//derive the node specification from that.
		final NodeSpecification nodeSpec = new NodeSpecification(keywordSpecs);
		//than create the sequences (one Bubble- and one Insertionsort).
		final Sequence a = new Sequence(nodeSpec);
		Node node = new Node(a);
		node.setValue("type", new SymbolicValue(nodeTypes, "class"));
		node.setValue("name", new StringValue("BubbleSort"));
		node.setValue("codePosition", new VectorialValue(new double[]{0, 10}));
		a.getNodes().add(node);
		node = new Node(a);
		node.setValue("type", new SymbolicValue(nodeTypes, "method"));
		node.setValue("name", new StringValue("sort"));
		node.setValue("codePosition", new VectorialValue(new double[]{1, 9}));
		a.getNodes().add(node);

		final Sequence b = new Sequence(nodeSpec);
		node = new Node(b);
		node.setValue("type", new SymbolicValue(nodeTypes, "class"));
		node.setValue("name", new StringValue("InsertionSort"));
		node.setValue("codePosition", new VectorialValue(new double[]{0, 10}));
		b.getNodes().add(node);
		node = new Node(b);
		node.setValue("type", new SymbolicValue(nodeTypes, "method"));
		node.setValue("name", new StringValue("sort"));
		node.setValue("codePosition", new VectorialValue(new double[]{1, 4}));
		b.getNodes().add(node);
		node = new Node(b);
		node.setValue("type", new SymbolicValue(nodeTypes, "method"));
		node.setValue("name", new StringValue("insert"));
		node.setValue("codePosition", new VectorialValue(new double[]{5, 9}));
		b.getNodes().add(node);

		//then specify alignment
		final AlignmentSpecification alignmentSpec = new AlignmentSpecification(nodeSpec);
		//set the weighting (node type is most important, name slightly and code position not at all).
		alignmentSpec.setWeighting(new double[]{0.9, 0.1, 0});

		//and finally do it.
		final StrictAlignmentFullAlgorithm algo = new StrictAlignmentFullAlgorithm(alignmentSpec);
		final AlignmentPath path = algo.calculateAlignment(a, b);
		assertEquals(0.67 / ((double) (a.getNodes().size() + b.getNodes().size())), path.getScore(),
				0.1);

		System.out.println(path.toString());
		//assert the exact path.
		assertEquals(3, path.getOperations().size());

		Operation op = path.getOperations().get(0);
		assertSame(a.getNodes().get(0), op.getLeft());
		assertSame(b.getNodes().get(0), op.getRight());

		op = path.getOperations().get(1);
		assertSame(a.getNodes().get(1), op.getLeft());
		assertSame(b.getNodes().get(1), op.getRight());

		op = path.getOperations().get(2);
		assertSame(null, op.getLeft());
		assertSame(b.getNodes().get(2), op.getRight());
	}

	public void testWeightLearning() {
		/*
		 * This is a toy example for the learning of weights using gradient
		 * descent on the non-normalized GLVQ cost function: E(w) = d^+(w) -
		 * d^-(w)
		 *
		 * We only consider one datapoint and two prototypes. All these
		 * sequences have only one node that models an excerpt from a java
		 * program. The weighting that shall be learned is that variable names
		 * and code position do not matter but the node type does.
		 */
		//specify the keywords.
		final KeywordSpecification[] keywordSpecs = new KeywordSpecification[3];
		//first the node type.
		final Alphabet nodeTypes = new Alphabet(new String[]{"for", "while", "switch"});
		keywordSpecs[0] = new SymbolicKeywordSpecification(nodeTypes, "type");
		//then the name
		keywordSpecs[1] = new StringKeywordSpecification("name");
		//then the code position (only start line and end line)
		keywordSpecs[2] = new VectorialKeywordSpecification(2, "codePosition");
		//derive the node specification from that.
		final NodeSpecification nodeSpec = new NodeSpecification(keywordSpecs);
		//than create the sequences

		//the actual datapoint.
		final Sequence a = new Sequence(nodeSpec);
		Node node = new Node(a);
		node.setValue("type", new SymbolicValue(nodeTypes, "for"));
		node.setValue("name", new StringValue("i"));
		node.setValue("codePosition", new VectorialValue(new double[]{10, 10}));
		a.getNodes().add(node);

		//the next positive prototype.
		final Sequence pPlus = new Sequence(nodeSpec);
		node = new Node(pPlus);
		node.setValue("type", new SymbolicValue(nodeTypes, "for"));
		node.setValue("name", new StringValue("j"));
		node.setValue("codePosition", new VectorialValue(new double[]{7, 7}));
		pPlus.getNodes().add(node);

		//the next negative prototype
		final Sequence pMinus = new Sequence(nodeSpec);
		node = new Node(pMinus);
		node.setValue("type", new SymbolicValue(nodeTypes, "while"));
		node.setValue("name", new StringValue("i"));
		node.setValue("codePosition", new VectorialValue(new double[]{8, 8}));
		pMinus.getNodes().add(node);

		//create the alignment algorithm.
		final AlignmentSpecification alignmentSpec = new AlignmentSpecification(nodeSpec);
		final StrictAlignmentFullAlgorithm algo = new StrictAlignmentFullAlgorithm(alignmentSpec);

		//no do some training runs and do a gradient descent.
		final double eta = 0.2;
		final int T = 10;
		for (int t = 0; t < T; t++) {
			final AlignmentPath pPath = algo.calculateAlignment(a, pPlus);
			final AlignmentPath mPath = algo.calculateAlignment(a, pMinus);

			final double[] pDerivative = pPath.calculateWeightDerivative();
			final double[] mDerivative = mPath.calculateWeightDerivative();

			//learn new weights.
			final double[] newWeights = new double[alignmentSpec.size()];
			double sum = 0;
			for (int k = 0; k < newWeights.length; k++) {
				newWeights[k] = alignmentSpec.getWeighting()[k]
						- eta * (pDerivative[k] - mDerivative[k]);
				//clip the weights at 0.
				if (newWeights[k] < 0) {
					newWeights[k] = 0;
				}
				sum += newWeights[k];
			}
			//normalize the weights.
			for (int k = 0; k < newWeights.length; k++) {
				newWeights[k] /= sum;
			}
			System.out.println(Arrays.toString(newWeights));
			//then set them
			alignmentSpec.setWeighting(newWeights);
		}
		assertEquals(1, alignmentSpec.getWeighting()[0], 0.1);
		assertEquals(0, alignmentSpec.getWeighting()[1], 0.1);
		assertEquals(0, alignmentSpec.getWeighting()[2], 0.1);

	}

	public void testReplacementLearning() {
		/*
		 * This is a toy example for the learning of replacement costs using
		 * gradient descent on the non-normalized GLVQ cost function: E(w) =
		 * d^+(w) - d^-(w)
		 *
		 * We only consider one datapoint and two prototypes. All these
		 * sequences have only one node that models an excerpt from a java
		 * program. We want to learn that "for" and "while" loops are the same
		 * while "switch" is something different.
		 */
		//specify the keywords.
		final KeywordSpecification[] keywordSpecs = new KeywordSpecification[1];
		//first the node type.
		final Alphabet nodeTypes = new Alphabet(new String[]{"for", "while", "switch"});
		keywordSpecs[0] = new SymbolicKeywordSpecification(nodeTypes, "type");
		//derive the node specification from that.
		final NodeSpecification nodeSpec = new NodeSpecification(keywordSpecs);
		//than create the sequences

		//the actual datapoint.
		final Sequence a = new Sequence(nodeSpec);
		Node node = new Node(a);
		node.setValue("type", new SymbolicValue(nodeTypes, "for"));
		a.getNodes().add(node);

		//the next positive prototype.
		final Sequence pPlus = new Sequence(nodeSpec);
		node = new Node(pPlus);
		node.setValue("type", new SymbolicValue(nodeTypes, "while"));
		pPlus.getNodes().add(node);

		//the next negative prototype
		final Sequence pMinus = new Sequence(nodeSpec);
		node = new Node(pMinus);
		node.setValue("type", new SymbolicValue(nodeTypes, "switch"));
		pMinus.getNodes().add(node);

		//create the alignment algorithm.
		final AlignmentSpecification alignmentSpec = new AlignmentSpecification(nodeSpec);
		final StrictAlignmentFullAlgorithm algo = new StrictAlignmentFullAlgorithm(alignmentSpec);
		//retrieve the replacement comparator.
		final ReplacementComparator comparator
				= (ReplacementComparator) alignmentSpec.getComparator("type");

		//no do some training runs and do a gradient descent.
		final double eta = 0.2;
		final int T = 4;
		final double[][] costs = comparator.getCostMatrix().getCostMatrix();
		for (int t = 0; t < T; t++) {
			final AlignmentPath pPath = algo.calculateAlignment(a, pPlus);
			final AlignmentPath mPath = algo.calculateAlignment(a, pMinus);

			final double[][] pDerivative = pPath.calculateParameterDerivative(
					comparator, "type");
			final double[][] mDerivative = mPath.calculateParameterDerivative(
					comparator, "type");

			//learn new replacement costs.
			double max = 0;
			for (int i = 0; i < costs.length; i++) {
				for (int j = 0; j < costs.length; j++) {
					costs[i][j] = costs[i][j]
							- eta * (pDerivative[i][j] - mDerivative[i][j]);
					//clip the weights at 0.
					if (costs[i][j] < 0) {
						costs[i][j] = 0;
					} else if (costs[i][j] > max) {
						max = costs[i][j];
					}
				}
			}
			//normalize the new parameters. 
			for (int i = 0; i < costs.length; i++) {
				for (int j = 0; j < costs.length; j++) {
					costs[i][j] /= max;
				}
			}
			System.out.println("for -> while : " + costs[0][1]
					+ "; for -> switch : " + costs[0][2]);
			//then set them
			comparator.getCostMatrix().setCostMatrix(costs);
		}
		//check the result.
		for (int i = 0; i < nodeTypes.size(); i++) {
			for (int j = 0; j < nodeTypes.size(); j++) {
				if (i == j) {
					assertEquals(0, costs[i][j], 0);
				} else if (i == 0 && j == 1) {
					//replacement of for -> while should cost nothing.
					assertEquals(0, costs[i][j], 0);
				} else if (i == 0 && j == 2) {
					//reeplacement of for -> switch should cost 1.
					assertEquals(1, costs[i][j], 0);

				} else {
					//all other replacement costs are non-discriminative and are
					//lowered by normalization.
					assertTrue(costs[i][j] > 0);
					assertTrue(costs[i][j] < 1);
				}
			}
		}
	}
}
