/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.comparators.ReplacementCosts;
import de.unibi.citec.tcs.alignment.sequence.Alphabet;
import de.unibi.citec.tcs.alignment.sequence.KeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.Node;
import de.unibi.citec.tcs.alignment.sequence.NodeSpecification;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.sequence.StringKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.StringValue;
import de.unibi.citec.tcs.alignment.sequence.SymbolicKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.SymbolicValue;
import de.unibi.citec.tcs.alignment.sequence.VectorialKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.VectorialValue;
import junit.framework.TestCase;

/**
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class StrictAlignmentScoreAlgorithmTest extends TestCase {

	public StrictAlignmentScoreAlgorithmTest(String testName) {
		super(testName);
	}

	@Override
	protected void setUp() throws Exception {
		super.setUp();
	}

	@Override
	protected void tearDown() throws Exception {
		super.tearDown();
	}

	/**
	 * Test of calculateAlignment method, of class StrictAlignmentAlgorithm.
	 */
	public void testAlignmentPerformance() {

		//align a...a and a...b
		final int length = 1000;

		final StringBuilder aBuilder = new StringBuilder();
		final StringBuilder bBuilder = new StringBuilder();

		aBuilder.append('b');
		bBuilder.append('b');
		for (int i = 1; i < length - 1; i++) {
			aBuilder.append('a');
			bBuilder.append('a');
		}
		aBuilder.append('a');
		bBuilder.append('b');

		final Alphabet alpha = new Alphabet("a|b");

		final Sequence a = new Sequence(aBuilder.toString(), alpha);
		final Sequence b = new Sequence(bBuilder.toString(), alpha);

		//initialize actual instance.
		final StrictAlignmentScoreAlgorithm instance
				= new StrictAlignmentScoreAlgorithm(new AlignmentSpecification(a));

		double expResult = 5E-4;
		final long start = System.nanoTime();
		final double result = instance.calculateAlignment(a, b);
		final double time = (double) (System.nanoTime() - start) / 1E9;
		assertEquals(expResult, result, 0);
	}

	public void testCalculateAlignment() {
		/*
		 * This is a test for an alignment of full blown sequences. Our test
		 * sequences here represent
		 * very simple java programs.
		 */
		//specify the keywords.
		final KeywordSpecification[] keywordSpecs = new KeywordSpecification[3];
		//first the node type.
		final Alphabet nodeTypes = new Alphabet(new String[]{"class", "method"});
		keywordSpecs[0] = new SymbolicKeywordSpecification(nodeTypes, "type");
		//then the name
		keywordSpecs[1] = new StringKeywordSpecification("name");
		//then the code position (only start line and end line)
		keywordSpecs[2] = new VectorialKeywordSpecification(2, "codePosition");
		//derive the node specification from that.
		final NodeSpecification nodeSpec = new NodeSpecification(keywordSpecs);
		//than create the sequences (one Bubble- and one Insertionsort).
		final Sequence a = new Sequence(nodeSpec);
		Node node = new Node(a);
		node.setValue("type", new SymbolicValue(nodeTypes, "class"));
		node.setValue("name", new StringValue("BubbleSort"));
		node.setValue("codePosition", new VectorialValue(new double[]{0, 10}));
		a.getNodes().add(node);
		node = new Node(a);
		node.setValue("type", new SymbolicValue(nodeTypes, "method"));
		node.setValue("name", new StringValue("sort"));
		node.setValue("codePosition", new VectorialValue(new double[]{1, 9}));
		a.getNodes().add(node);

		final Sequence b = new Sequence(nodeSpec);
		node = new Node(b);
		node.setValue("type", new SymbolicValue(nodeTypes, "class"));
		node.setValue("name", new StringValue("InsertionSort"));
		node.setValue("codePosition", new VectorialValue(new double[]{0, 10}));
		b.getNodes().add(node);
		node = new Node(b);
		node.setValue("type", new SymbolicValue(nodeTypes, "method"));
		node.setValue("name", new StringValue("sort"));
		node.setValue("codePosition", new VectorialValue(new double[]{1, 4}));
		b.getNodes().add(node);
		node = new Node(b);
		node.setValue("type", new SymbolicValue(nodeTypes, "method"));
		node.setValue("name", new StringValue("insert"));
		node.setValue("codePosition", new VectorialValue(new double[]{5, 9}));
		b.getNodes().add(node);

		//then specify alignment
		final AlignmentSpecification alignmentSpec = new AlignmentSpecification(nodeSpec);
		//ignore two keywords using the weighting.
		alignmentSpec.setWeighting(new double[]{1, 0, 0});

		//and finally do it.
		final StrictAlignmentScoreAlgorithm algo = new StrictAlignmentScoreAlgorithm(alignmentSpec);
		final double score = algo.calculateAlignment(a, b);
		assertEquals(ReplacementCosts.GAPDEFAULT / ((double) (a.getNodes().size() + b.getNodes().size())), score, 0);
	}
}
