/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment.sequence;

import de.unibi.citec.tcs.alignment.comparators.ReplacementCosts;
import java.util.Arrays;
import junit.framework.TestCase;

/**
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class ReplacementCostsTest extends TestCase {

	private Alphabet testAlpha;
	private ReplacementCosts testMatrix;

	public ReplacementCostsTest(String testName) {
		super(testName);
	}

	@Override
	protected void setUp() throws Exception {
		super.setUp();
		testAlpha = new Alphabet("ab");
		testMatrix = new ReplacementCosts(testAlpha, true, true);
	}

	@Override
	protected void tearDown() throws Exception {
		super.tearDown();
		testAlpha = null;
		testMatrix = null;
	}

	/**
	 * Test of isAlignmentMatrix method, of class ReplacementCosts.
	 */
	public void testIsAlignmentMatrix() {
		assertTrue(testMatrix.isGapAlignmentMatrix());
	}

	/**
	 * Test of isLocalAlignmentMatrix method, of class ReplacementCosts.
	 */
	public void testIsLocalAlignmentMatrix() {
		assertTrue(testMatrix.isLocalAlignmentMatrix());
	}

	/**
	 * Test of getCostMatrix method, of class ReplacementCosts.
	 */
	public void testGetCostMatrix() {
		//test defaults.
		final double[][] costMatrix = testMatrix.getCostMatrix();
		assertEquals(testAlpha.size() + 2, costMatrix.length);
		for (int i = 0; i < costMatrix.length; i++) {
			assertEquals(testAlpha.size() + 2, costMatrix[i].length);
			for (int j = 0; j < costMatrix.length; j++) {
				if (i == j) {
					assertEquals(0, costMatrix[i][j], 0);
				} else if ((i == testAlpha.size() && j < testAlpha.size()) ||
						(i < testAlpha.size() && j == testAlpha.size())) {
					assertEquals(ReplacementCosts.GAPDEFAULT, costMatrix[i][j], 0);
				} else if ((i == testAlpha.size() + 1 && j < testAlpha.size()) ||
						(i < testAlpha.size() && j == testAlpha.size() + 1)) {
					assertEquals(ReplacementCosts.SKIPDEFAULT, costMatrix[i][j], 0);
				} else if (i < testAlpha.size() && j < testAlpha.size()) {
					assertEquals(ReplacementCosts.MISMATCHDEFAULT, costMatrix[i][j], 0);
				}
			}
		}
		//also check if changes influence the inner state (which should be the case here)
		costMatrix[0][1] = 5;
		assertEquals(5, testMatrix.getCostMatrix()[0][1], 0);
	}

	/**
	 * Test of setCostMatrix method, of class ReplacementCosts.
	 */
	public void testSetCostMatrix() {
		final double[][] costMatrix = {{0, 3, 2, 1}, {3, 0, 2, 1}, {2, 2, 0, 0}, {1, 1, 0, 0}};
		testMatrix.setCostMatrix(costMatrix);
		assertTrue(Arrays.deepEquals(costMatrix, testMatrix.getCostMatrix()));
	}

	/**
	 * Test of getReplacementCosts method, of class ReplacementCosts.
	 */
	public void testGetReplacementCosts() {
		final double[][] expected = {{0, ReplacementCosts.MISMATCHDEFAULT}, {
				ReplacementCosts.MISMATCHDEFAULT, 0}};
		assertTrue(Arrays.deepEquals(expected, testMatrix.getReplacementCosts()));
	}

	/**
	 * Test of setReplacementCosts method, of class ReplacementCosts.
	 */
	public void testSetReplacementCosts() {
		final double[][] repCosts = {{0, 3}, {3, 0}};
		testMatrix.setReplacementCosts(repCosts);
		assertTrue(Arrays.deepEquals(repCosts, testMatrix.getReplacementCosts()));
	}

	/**
	 * Test of getReplacementCost method, of class ReplacementCosts.
	 */
	public void testGetReplacementCost() {
		assertEquals(0, testMatrix.getReplacementCost("a", "a"), 0);
		assertEquals(ReplacementCosts.MISMATCHDEFAULT, testMatrix.getReplacementCost("a", "b"), 0);
		assertEquals(ReplacementCosts.MISMATCHDEFAULT, testMatrix.getReplacementCost("b", "a"), 0);
		assertEquals(0, testMatrix.getReplacementCost("b", "b"), 0);
	}

	/**
	 * Test of setReplacementCost method, of class ReplacementCosts.
	 */
	public void testSetReplacementCost() {
		testMatrix.setReplacementCost("a", "b", 3);
		assertEquals(3, testMatrix.getReplacementCost("a", "b"), 0);
	}

	/**
	 * Test of getDeletionCosts method, of class ReplacementCosts.
	 */
	public void testGetDeletionCosts() {
		final double[] expected = {ReplacementCosts.GAPDEFAULT, ReplacementCosts.GAPDEFAULT};
		assertTrue(Arrays.equals(expected, testMatrix.getDeletionCosts()));
	}

	/**
	 * Test of setDeletionCosts method, of class ReplacementCosts.
	 */
	public void testSetDeletionCosts() {
		final double[] delCosts = {3, 1};
		testMatrix.setDeletionCosts(delCosts);
		assertTrue(Arrays.equals(delCosts, testMatrix.getDeletionCosts()));
	}

	/**
	 * Test of getDeletionCost method, of class ReplacementCosts.
	 */
	public void testGetDeletionCost() {
		assertEquals(ReplacementCosts.GAPDEFAULT, testMatrix.getDeletionCost("a"), 0);
		assertEquals(ReplacementCosts.GAPDEFAULT, testMatrix.getDeletionCost("b"), 0);
	}

	/**
	 * Test of setDeletionCost method, of class ReplacementCosts.
	 */
	public void testSetDeletionCost() {
		testMatrix.setDeletionCost("a", 5);
		assertEquals(5, testMatrix.getDeletionCost("a"), 0);
	}

	/**
	 * Test of getInsertionCosts method, of class ReplacementCosts.
	 */
	public void testGetInsertionCosts() {
		final double[] expected = {ReplacementCosts.GAPDEFAULT, ReplacementCosts.GAPDEFAULT};
		assertTrue(Arrays.equals(expected, testMatrix.getInsertionCosts()));
	}

	/**
	 * Test of setInsertionCosts method, of class ReplacementCosts.
	 */
	public void testSetInsertionCosts() {
		final double[] insCosts = {3, 1};
		testMatrix.setInsertionCosts(insCosts);
		assertTrue(Arrays.equals(insCosts, testMatrix.getInsertionCosts()));
	}

	/**
	 * Test of getInsertionCost method, of class ReplacementCosts.
	 */
	public void testGetInsertionCost() {
		assertEquals(ReplacementCosts.GAPDEFAULT, testMatrix.getInsertionCost("a"), 0);
		assertEquals(ReplacementCosts.GAPDEFAULT, testMatrix.getInsertionCost("b"), 0);
	}

	/**
	 * Test of setInsertionCost method, of class ReplacementCosts.
	 */
	public void testSetInsertionCost() {
		testMatrix.setInsertionCost("a", 5);
		assertEquals(5, testMatrix.getInsertionCost("a"), 0);
	}

	/**
	 * Test of getSkipDeletionCosts method, of class ReplacementCosts.
	 */
	public void testGetSkipDeletionCosts() {
		final double[] expected = {ReplacementCosts.SKIPDEFAULT, ReplacementCosts.SKIPDEFAULT};
		assertTrue(Arrays.equals(expected, testMatrix.getSkipDeletionCosts()));
	}

	/**
	 * Test of setSkipDeletionCosts method, of class ReplacementCosts.
	 */
	public void testSetSkipDeletionCosts() {
		final double[] delCosts = {3, 1};
		testMatrix.setSkipDeletionCosts(delCosts);
		assertTrue(Arrays.equals(delCosts, testMatrix.getSkipDeletionCosts()));
	}

	/**
	 * Test of getSkipDeletionCost method, of class ReplacementCosts.
	 */
	public void testGetSkipDeletionCost() {
		assertEquals(ReplacementCosts.SKIPDEFAULT, testMatrix.getSkipDeletionCost("a"), 0);
		assertEquals(ReplacementCosts.SKIPDEFAULT, testMatrix.getSkipDeletionCost("b"), 0);
	}

	/**
	 * Test of setSkipDeletionCost method, of class ReplacementCosts.
	 */
	public void testSetSkipDeletionCost() {
		testMatrix.setSkipDeletionCost("a", 4);
		assertEquals(4, testMatrix.getSkipDeletionCost("a"), 0);
	}

	/**
	 * Test of getSkipInsertionCosts method, of class ReplacementCosts.
	 */
	public void testGetSkipInsertionCosts() {
		final double[] expected = {ReplacementCosts.SKIPDEFAULT, ReplacementCosts.SKIPDEFAULT};
		assertTrue(Arrays.equals(expected, testMatrix.getSkipInsertionCosts()));
	}

	/**
	 * Test of setSkipInsertionCosts method, of class ReplacementCosts.
	 */
	public void testSetSkipInsertionCosts() {
		final double[] insCosts = {3, 1};
		testMatrix.setSkipInsertionCosts(insCosts);
		assertTrue(Arrays.equals(insCosts, testMatrix.getSkipInsertionCosts()));
	}

	/**
	 * Test of getSkipInsertionCost method, of class ReplacementCosts.
	 */
	public void testGetSkipInsertionCost() {
		assertEquals(ReplacementCosts.SKIPDEFAULT, testMatrix.getSkipInsertionCost("a"), 0);
		assertEquals(ReplacementCosts.SKIPDEFAULT, testMatrix.getSkipInsertionCost("b"), 0);
	}

	/**
	 * Test of setSkipInsertionCost method, of class ReplacementCosts.
	 */
	public void testSetSkipInsertionCost() {
		testMatrix.setSkipInsertionCost("a", 4);
		assertEquals(4, testMatrix.getSkipInsertionCost("a"), 0);
	}
}
